import numpy as np

from .data_transform import DataTransformer

class FactorizationMachine:
    """
    A Factorization Machine class 
    
    """
    
    def __init__(self, users, items, y, v_matrix):
        self.users = users
        self.items = items
        self.n_users = len(users)
        self.n_items = len(items)
        self.y = y
        
        self.data_transformer = DataTransformer(users, items, y)
        self.training_data, self.training_y = self.data_transformer.transform()
        
        self.w_bias = np.random.normal()
        self.w_pars = np.random.normal(size=self.n_users + self.n_items)
        self.v_matrix = np.copy(v_matrix)
        
        
    
    def predict(self, user_idx, item_idx):
        n = self.training_data.shape[0]
        n_users = len(self.users)
        n_items = len(self.items)
        
        x_hat = self.w_bias + self.w_pars[user_idx] + self.w_pars[n_users + item_idx] + self.v_matrix[user_idx, :]@self.v_matrix[user_idx + item_idx, :].T
        
        return x_hat

    def predict_all(self):
        n = self.training_data.shape[0]
        n_users = len(self.users)
        n_items = len(self.items)

        X_hat = np.zeros_like(self.y)
        for user_idx in range(n_users):
            for item_idx in range(n_items):
                x_hat = self.predict(user_idx, item_idx)
    
                X_hat[user_idx, item_idx] = x_hat

        return X_hat
    
    def loss_MSE(self):
        
        y_hats = []
        
        for user_idx in range(len(self.users)):
            for item_idx in range(len(self.items)):
                y_hat = self.predict(user_idx, item_idx)
                y_hats.append(y_hat)
               
            
        y_hats = np.array(y_hats)
               
        mse = np.mean((y_hats - self.training_y)**2)
            
        return 1/2 * mse
    
    
    def loss_MSE_gradient(self, x_input):
        n_features = self.v_matrix.shape[1]
        n = self.v_matrix.shape[0]
        
        w_bias_gradient = -1
        w_pars_gradient = -x_input
        
        v_gradients = np.zeros_like(self.v_matrix)

        for idx_i in range(n):
            x_i = x_input[idx_i]
            x_i_vector = np.array([x_i**2] * n)
            
            for feature_idx in range(n_features):
                
                grad_sum = np.dot(self.v_matrix[:, feature_idx], x_input) - np.dot(self.v_matrix[:, feature_idx], x_i_vector)
                
                v_grad = x_i * grad_sum
                v_gradients[idx_i, feature_idx] = -v_grad
        
        
        
        
        return w_bias_gradient, w_pars_gradient, v_gradients


from collections import Counter
import pandas as pd

def load_knowledge_base_triples(csv_path, cols):
    """Function for loading knowledge base triples from csv file
    
    Parameters
    ----------
    csv_path : path to csv file
    cols : subject, predicate and object column names

    Returns
    -------
    tuple of parsed knowledge base triples and new knowledge base
    """

    knowledge_base_data = pd.read_csv(csv_path)
    knowledge_base_data["genre"] = "genre"
    knowledge_base_data = knowledge_base_data[["movie", "genre", "subject"]]

    knowledge_base_sparql_triples = knowledge_base_data.values

    a = [ (predicate, obj) for (item, predicate, obj) in knowledge_base_sparql_triples ]
    ctr = Counter(a)
    features_only = [ feature for (feature, _) in ctr.most_common(25) ]
    knowledge_base_new = [ (subj, pred, obj) for (subj, pred, obj) in knowledge_base_sparql_triples if (pred, obj) in features_only ]

    return knowledge_base_sparql_triples, knowledge_base_new


import numpy as np

class StochasticGradientDescent:
    
    def __init__(self, factorization_machine, iterations=500, learning_rate=0.0001):
        self.factorization_machine = factorization_machine
        self.learning_rate = learning_rate
        self.iterations = iterations
        
        
    def fit(self):
        training_data = self.factorization_machine.training_data
        
        loss_data = []
        
        
        for i in range(self.iterations):
            print("iter %d" % i)
            bias_gradient = 0
            w_gradient = np.zeros_like(self.factorization_machine.w_pars)
            v_gradient = np.zeros_like(self.factorization_machine.v_matrix)
        
        
            for row_idx in range(training_data.shape[0]):
                current_bias_gradient, current_pars_gradient, current_v_gradient = self.factorization_machine.loss_MSE_gradient(training_data[row_idx, :])

                bias_gradient += current_bias_gradient
                w_gradient += current_pars_gradient
                v_gradient += current_v_gradient
            
            
            loss_MSE = self.factorization_machine.loss_MSE()
            bias_gradient *= loss_MSE
            w_gradient *= loss_MSE
            v_gradient *= loss_MSE
            
            print("loss MSE", loss_MSE)
            loss_data.append(loss_MSE)
            
            self.factorization_machine.w_bias -= self.learning_rate * bias_gradient
            self.factorization_machine.w_pars -= self.learning_rate * w_gradient
            self.factorization_machine.v_matrix -= self.learning_rate * v_gradient
            
            
        return loss_data
            
        


import numpy as np


class DataTransformer:
    """DataTransformer is a utility class for transforming
    user item matrix to a data structure more suited for a
    factorization machine
    

    Parameters
    ----------
    users: array of users
    items: array of items
    x_matrix: user items matrix

    """

    def __init__(self, users, items, x_matrix):
        self.users = users
        self.items = items
        self.x_matrix = x_matrix

    def transform(self):
        n_users = len(self.users)
        n_items = len(self.items)

        target_matrix = np.zeros(shape=(n_users * n_items, n_users + n_items))
        target_matrix_y = np.zeros(n_users * n_items)

        row_counter = 0
        for user_idx, user in enumerate(self.users):
            for item_idx, item in enumerate(self.items):
                target_matrix[row_counter, user_idx] = 1
                target_matrix[row_counter, (n_users - 1) + item_idx] = 1

                target_matrix_y[row_counter] = self.x_matrix[user_idx, item_idx]
                
                row_counter += 1


        self.data_x = target_matrix
        self.data_y = target_matrix_y

        return self.data_x, self.data_y


class TFIDFTransformer:
    
    def __init__(self, knowledge_base_triples, x_matrix, user_names, item_names):
        
        self.x_matrix = x_matrix
        
        self.user_names = user_names
        self.item_names = item_names
        self.n_users, self.n_items = x_matrix.shape
        self.knowledge_base_triples = knowledge_base_triples
        self.all_possible_features_F = list({ (predicate, obj) for
                                             (item, predicate, obj)
                                             in knowledge_base_triples })
        
        self.v_matrix = None
        self.v_matrix_users = None
        self.v_matrix_items = None
        
        
    def generate_v_for_item(self, current_item, feature):
        feature_predicate, feature_object = feature
        tf_numerator = 1 if (current_item, feature_predicate, feature_object) in self.knowledge_base_triples else 0

        tf_denominator = 0
        for current_predicate, current_obj in self.all_possible_features_F:
            predicates = { (predicate, obj) for (item, predicate, obj) in self.knowledge_base_triples
                         if current_item == item and current_predicate == predicate and current_obj == obj }
            n_predicates = len(predicates)

            tf_denominator += n_predicates**2

        tf_denominator = np.sqrt(tf_denominator)

        idf_numerator = self.n_items
        idf_denominator = len({ item for (item, predicate, obj) in self.knowledge_base_triples
                              if feature_predicate == predicate and feature_object == obj })

        v_value = tf_numerator / tf_denominator * np.log(idf_numerator / idf_denominator)

        return v_value
    
    
    def generate_v_for_user(self, user, feature):
        feature_idx = self.all_possible_features_F.index(feature)
        user_idx = self.user_names.index(user)

        items_enjoyed_by_user = [ item_idx 
                                 for item_idx, user_enjoyment 
                                 in enumerate(self.x_matrix[user_idx, :])
                                 if user_enjoyment == 1 ]    

        u_numerator = np.sum(self.v_matrix_items[items_enjoyed_by_user, feature_idx])
        u_denominator = len({ movie_idx for movie_idx in items_enjoyed_by_user
                            if self.v_matrix_items[movie_idx, feature_idx] != 0 })

        if u_denominator == 0:
            return u_denominator

        u_value = u_numerator / u_denominator

        return u_value
    
    
    def generate_v_item_matrix(self):
        v_matrix = np.zeros((self.n_items, len(self.all_possible_features_F)))

        for i, item in enumerate(self.item_names):
            for j, (predicate, obj) in enumerate(self.all_possible_features_F):
                v_matrix[i, j] = self.generate_v_for_item(item, (predicate, obj))

        self.v_matrix_items = v_matrix
        
        
    def generate_v_user_matrix(self):
        u_matrix = np.zeros((self.n_users, len(self.all_possible_features_F)))

        for i, user in enumerate(self.user_names):
            for j, (predicate, obj) in enumerate(self.all_possible_features_F):
                u_matrix[i, j] = self.generate_v_for_user(user, (predicate, obj))

        self.v_matrix_users = u_matrix
        
    def generate_v_matrix(self):
        self.generate_v_item_matrix()
        self.generate_v_user_matrix()
        
        self.v_matrix = np.concatenate([self.v_matrix_users, self.v_matrix_items])


from .data_transform import TFIDFTransformer, DataTransformer
from .factor_machine import FactorizationMachine
from .gradient_descent import StochasticGradientDescent
from .data_loader import load_knowledge_base_triples
#from .knowledge_base import generate_knowledge_base




from setuptools import setup

from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
  name = 'pykaHFM',
  packages = ['pykaHFM'],
  version = '0.2.2',
  description = 'An implementation of kaHFM algorithm',
  author = 'Jiří Filip',
  author_email = 'j.f.ilip@seznam.cz',
  long_description=long_description,
  long_description_content_type='text/markdown',
  url = 'https://github.com/jirifilip/pykaHFM',
  keywords = 'pykaHFM',
  classifiers = [],
  install_requires=['pandas', 'numpy']
)


def generate_knowledge_base(items, predicates):
    knowledge_base = []
    
    for item in items:
        for predicate in predicates.keys():
            options = predicates[predicate]
            n_options = len(options)
            
            object_subset = np.random.choice(options, np.random.randint(1, n_options + 1))
            for chosen_object in object_subset:
                knowledge_base.append((item, predicate, chosen_object))
    return knowledge_base